"""
Leases management :
download from r2labapi service,
creation update deletions,
and minimal verifications
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w1202, r1705, w0703

import os
import pwd
import time
import traceback
from datetime import datetime, timezone

import requests

from .logger import logger
from .config import Config
from .r2labapiproxy import R2labApiProxy, iso_to_epoch, epoch_to_iso

DEBUG = False
DEBUG = True


class Lease:
    """
    a single lease entry
    """

    wire_timeformat = "%Y-%m-%dT%H:%M:%S%Z"

    def __init__(self, api_lease):
        try:
            self.owner = api_lease.get('slice_name') or '(unknown)'
            self.lease_id = api_lease['id']
            self.ifrom = iso_to_epoch(api_lease['t_from'])
            self.iuntil = iso_to_epoch(api_lease['t_until'])
            self.broken = False

        except Exception as exc:
            self.broken = f"lease broken b/c of exception {exc}"

    # show a 2-chars prefix
    # one for rights (computed by caller - not shown if not provided)
    # one for situation wrt now: < for past, = for current, > for future
    def __repr__(self, rights_char=None):
        rights_char = ('B' if self.broken
                       else rights_char if rights_char
                       else '?')
        now = time.time()
        if self.iuntil < now:
            time_char = '<'
        elif self.ifrom < now:
            time_char = '='
        else:
            time_char = '>'
        # show second day only when they are different
        second_day = not self.same_date(self.ifrom, self.iuntil)
        time_message = (f"from {self.human(self.ifrom, show_timezone=False)}"
                        f" until {self.human(self.iuntil, show_date=second_day)}")
        return (f"{rights_char}{time_char}"
                f" {time_message} {self.owner}")

    def sort_key(self):
        """
        Sort on start time
        """
        return self.ifrom

    @staticmethod
    def human(epoch, show_date=True, show_timezone=True):
        """
        human-readable format
        """
        timeformat = ""
        timeformat += "%H:%M"
        if show_date:
            timeformat += " [on %m-%d]"
        if show_timezone:
            timeformat += " %Z"
        return time.strftime(timeformat, time.localtime(epoch))

    @staticmethod
    def same_date(epoch1, epoch2):
        """
        Returns True if both instants belong in the same day
        for local timezone
        """
        return (
            time.strftime("%y-%m-%d", time.localtime(epoch1))
            ==
            time.strftime("%y-%m-%d", time.localtime(epoch2))
        )

    def booked_now_by(self, login):
        """
        tells if this lease is held right now, by that login (if not None)
        or by anyone (if login is None)
        """
        return self.booked_at_by(time.time(), login)

    def booked_at_by(self, instant, login):
        """
        tells if this lease is held at that time, by that login (if not None)
        or by anyone (if login is None)
        """
        if self.broken:
            if DEBUG:
                logger.info(f"ignoring broken lease {self}")
            return False
        if not self.ifrom <= instant <= self.iuntil:
            if DEBUG:
                logger.info(f"{self} : wrong timerange")
            return False
        if login is not None and not self.owner == login:
            if DEBUG:
                logger.info(
                    f"login {login} is not owner - actual owner is {self.owner}")
            return False
        return self


class Leases:                                           # pylint: disable=r0902
    """
    A list of leases as downloaded from the API
    """

    def __init__(self, message_bus):
        self.message_bus = message_bus
        # don't use os.getlogin() as this gives root if under su
        self.login = pwd.getpwuid(os.getuid())[0]
        the_config = Config()
        api_url = the_config.value('r2labapi', 'url')
        self.resource_name = the_config.value('r2labapi', 'resource_name')
        # no token: reads are public, writes will prompt for credentials
        self.proxy = R2labApiProxy(api_url)
        # computed later
        # a list of Lease objects
        self.leases = None
        # the raw API response
        self.api_leases = None
        # xxx this is still used by monitornodes
        # should be cleaned up
        self.resources = None

    def __repr__(self):
        if self.leases is None:
            return (f"<Leases from {self.proxy} - **(UNFETCHED)**>")
        else:
            return (f"<Leases from {self.proxy}"
                    f" - {len(self.leases)} lease(s)>")

    async def feedback(self, field, msg):
        """
        send feedback, for displaying or monitoring
        """
        await self.message_bus.put({field: msg})

    def has_special_privileges(self):
        """
        check for being run as root
        """
        privileged = Config().value('accounts', 'privileged')
        names = privileged.split(',')
        return self.login in names

    async def booked_now_by_me(self, *, root_allowed=True):
        """
        fetch leases and return a bool that says if current login has a lease
        if root_allowed is True, then this function will return True
          for privileged users, no matter what else
        if root_allowed is false, privileged users go through the usual
          process, which most of the time has this function answering False
        """
        return await self.booked_now_by(login=self.login,
                                        root_allowed=root_allowed)

    async def booked_now_by(self, login, *, root_allowed=True):
        if root_allowed and self.has_special_privileges():
            return True
        try:
            await self.fetch_all()
            return self._booked_now_by_login(login)
        except Exception as exc:
            await self.feedback('info',
                                f"Could not fetch leases : {exc}")
            return False

    async def booked_now_by_anyone(self):
        """
        fetch leases and return a bool that says if a lease is currently valid
        """
        try:
            await self.fetch_all()
            return self._booked_now_by_anyone()
        except Exception as exc:
            await self.feedback('info', f"Could not fetch leases : {exc}")
            return False

    # the following 2 methods assume the leases have been fetched
    def _booked_now_by_login(self, login):
        # must have run fetch_all() before calling this
        return any([lease.booked_now_by(login)
                    for lease in self.leases])

    def _booked_now_by_anyone(self):
        # must have run fetch_all() before calling this
        return any([lease.booked_now_by(login=None)
                    for lease in self.leases])

    async def fetch_all(self):
        """
        makes sure all required data is available
        (turns out only the leases are required for now)
        """
        await self.fetch_leases()

    async def refresh(self):
        self.leases = None
        await self.fetch_all()

    def sort_leases(self):
        self.leases.sort(key=Lease.sort_key)

    async def fetch_leases(self):
        if self.leases is not None:
            return self.leases
        await self._fetch_leases()
        return self.leases

    # xxx stolen from r2lab.inria.fr : because we still use
    # a data model inspired from when we had OMF
    @staticmethod
    def epoch_to_ui_ts(epoch):
        return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(epoch))

    def resource_from_lease(self, api_lease):
        return {'uuid': api_lease['id'],
                'slicename': api_lease.get('slice_name', ''),
                'valid_from': self.epoch_to_ui_ts(
                    iso_to_epoch(api_lease['t_from'])),
                'valid_until': self.epoch_to_ui_ts(
                    iso_to_epoch(api_lease['t_until'])),
                'ok': True}

    async def _fetch_leases(self):
        self.leases = None
        try:
            logger.info("Leases are being fetched..")
            # fetch leases from today onwards
            today = datetime.now(tz=timezone.utc).replace(
                hour=0, minute=0, second=0, microsecond=0)
            self.api_leases = self.proxy.get_leases(
                after=today.isoformat())
            logger.info(f"{len(self.api_leases)} leases received")
            # decoded as a list of Lease objects
            self.leases = [Lease(entry) for entry in self.api_leases]
            self.sort_leases()
            self.resources = [
                self.resource_from_lease(api_lease)
                for api_lease in self.api_leases
            ]

        except requests.exceptions.RequestException as exc:
            if hasattr(exc, 'response') and exc.response is not None:
                message = (f"HTTP error ({exc.response.status_code})"
                           f" from {self.proxy}")
            else:
                message = f"cannot reach r2lab API at {self.proxy}: {exc}"
            logger.error(message)
            print(message)
            await self.feedback('leases_error', message)
        except Exception as exc:
            if DEBUG:
                print(f"Leases.fetch: exception {exc}")
            traceback.print_exc()
            await self.feedback(
                'leases_error',
                f"cannot get leases from {self} - exception {exc}")

    # this can be used with a fake message queue, it's synchroneous
    def print(self):
        print(5 * '-', self,
              "with special privileges"
              if self.has_special_privileges() else "")
        if self.leases is not None:
            def rights_char(lease):
                # B like broken
                if lease.broken:
                    return 'B'
                # S like super user
                if self.has_special_privileges():
                    return 'S'
                # * for yes you can use it
                if lease.booked_now_by(self.login):
                    return '*'
                return ' '
            for i, lease in enumerate(self.leases):
                msg = lease.__repr__(rights_char=rights_char(lease))
                print(f"{i+1:3d} {msg}")

    # material to create and modify leases
    @staticmethod
    def to_epoch(incoming):
        if not incoming:
            return time.time()
        if isinstance(incoming, (int, float)):
            return incoming
        patterns = [
            # fill in year
            "%Y-{}:00%Z",  # "%Y-{}:00",
            "%Y-%m-{}:00%Z",  # "%Y-%m-{}:00",
            "%Y-%m-%dT{}:00%Z",  # "%Y-%m-%dT{}:00",
            "%Y-%m-%dT{}:00:00%Z",  # "%Y-%m-%dT{}:%M:00",
        ]

        for pattern in patterns:
            fill = time.strftime(pattern).format(incoming)
            try:
                struct_time = time.strptime(fill, Lease.wire_timeformat)
                return int(time.mktime(struct_time))
            except Exception:
                pass

    @staticmethod
    def check_user(user):
        return os.path.exists(f"/home/{user}")

    async def _add_lease(self, owner, input_from, input_until):
        if owner != 'root':
            if not self.check_user(owner):
                print(f"user {owner} not found under /home - giving up")
                logger.error(f"Unknown user {owner}")
                return
        t_from = Leases.to_epoch(input_from)
        if not t_from:
            print(f"invalid time from: {input_from}")
            return
        t_until = Leases.to_epoch(input_until)
        if not t_until:
            print(f"invalid time until: {input_until}")
            return
        try:
            self.proxy.create_lease({
                'resource_name': self.resource_name,
                'slice_name': owner,
                't_from': epoch_to_iso(t_from),
                't_until': epoch_to_iso(t_until),
            })
            print("OK")
            # force next reload
            self.leases = None
        except Exception as exc:
            print('Error', f"Cannot add lease {type(exc)}: {exc}")
            traceback.print_exc()

    def get_lease_by_rank(self, lease_rank):
        try:
            irank = int(lease_rank)
            return self.leases[irank - 1]
        except Exception:
            pass

    async def _update_lease(self, lease_rank, input_from, input_until):
        update_fields = {}
        if not input_from and not input_until:
            logger.info("update_lease : nothing to do")
            return
        if input_from:
            t_from = Leases.to_epoch(input_from)
            if not t_from:
                print(f"invalid time from: {input_from}")
                return
            update_fields['t_from'] = epoch_to_iso(t_from)
        if input_until:
            t_until = Leases.to_epoch(input_until)
            if not t_until:
                print(f"invalid time until: {input_until}")
                return
            update_fields['t_until'] = epoch_to_iso(t_until)
        # lease_rank could be a rank as displayed by self.print()
        the_lease = self.get_lease_by_rank(lease_rank)
        if not the_lease:
            print(f"Cannot find lease with rank {lease_rank}")
            return
        try:
            self.proxy.update_lease(the_lease.lease_id, update_fields)
            print("OK")
            # force next reload
            self.leases = None
        except Exception as exc:
            print(f"error: {exc}")

    async def _delete_lease(self, lease_rank):
        # lease_rank could be a rank as displayed by self.print()
        the_lease = self.get_lease_by_rank(lease_rank)
        if not the_lease:
            print(f"Cannot find lease with rank {lease_rank}")
            return
        try:
            self.proxy.delete_lease(the_lease.lease_id)
            print("OK")
            # force next reload
            self.leases = None
        except Exception as exc:
            print(f"not deleted: {exc}")

    async def main(self, interactive):
        await self.fetch_all()
        self.print()
        if not interactive:
            return 0
        try:
            result = await self.interactive()
            return result
        except (KeyboardInterrupt, EOFError):
            print("Bye")
            return 1

    async def interactive(self):
        help_message = """
Enter one of the letters inside [], and answer the questions

A lease index is a number as shown on the left in the leases list

Times can be entered simply as
* just 14, or 14:00, for today at 2p.m.
* 14:30 for today at 2:30 p.m., or
* 27T10:30 for the 27th this month at 10:30 a.m., or
* 12-10T01:00 for the 12th december this year at 1 a.m., or
* 2016-01-02T08:00 for January 1st, 2016, at 8:00

In all the above cases, times will be understood as local time (France).

Leaving a time empty means either 'now', or 'do not change',
depending on the context
"""
        # interactive mode
        while True:
            current_time = time.strftime("%H:%M")
            answer = input(f"{current_time} - Enter command "
                           f"([l]ist, [a]dd, [u]pdate, "
                           f"[d]elete, [r]efresh, [q]uit : ")
            char = answer[0].lower() if answer else 'l'
            if char == 'l':
                self.print()
            elif char == 'a':
                if self.has_special_privileges():
                    owner = input("For slice name : ")
                else:
                    owner = self.login
                time_from = input("From : ")
                time_until = input("Until : ")
                print(f"owner={owner}")
                await self._add_lease(owner, time_from, time_until)
                await self.fetch_all()
            elif char == 'u':
                rank = input("Enter lease index : ")
                time_from = input("From : ")
                time_until = input("Until : ")
                await self._update_lease(rank, time_from, time_until)
                await self.fetch_all()
            elif char == 'd':
                rank = input("Enter lease index : ")
                await self._delete_lease(rank)
                await self.fetch_all()
            elif char == 'r':
                await self.refresh()
                self.print()
            elif char == 'q':
                print('bye')
                break
            else:
                print(help_message)
        return 0



"""
The Node object is the handle for doing most CMC-oriented actions like
querying status, turning on or off, and similar
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202
# pylint: disable=logging-fstring-interpolation

import os.path
import traceback

import asyncio
import aiohttp

from rhubarbe.logger import logger
from rhubarbe.config import Config
from rhubarbe.inventorynodes import InventoryNodes
from rhubarbe.frisbee import Frisbee
from rhubarbe.imagezip import ImageZip


class Node:                                             # pylint: disable=r0902

    """
    This class allows to talk to various parts of a node
    created from the cmc hostname for convenience
    the inventory lets us spot the other parts (control, essentially)
    """
    def __init__(self, cmc_name, message_bus):
        self.cmc_name = cmc_name
        self.message_bus = message_bus
        self.status = None
        self.action = None
        self.mac = None
        # for monitornodes
        self.id = int("".join([x for x in cmc_name      # pylint: disable=c0103
                               if x in "0123456789"]))
        # used later
        self.frisbee = None
        self.imagezip = None

    def __repr__(self):
        return f"<Node {self.control_hostname()}>"

    def is_known(self):
        return self.control_mac_address() is not None

    def control_mac_address(self):
        the_inventory = InventoryNodes()
        return the_inventory.attached_hostname_info(self.cmc_name,
                                                    'control', 'mac')

    def control_ip_address(self):
        the_inventory = InventoryNodes()
        return the_inventory.attached_hostname_info(self.cmc_name,
                                                    'control', 'ip')

    def control_hostname(self):
        the_inventory = InventoryNodes()
        return the_inventory.attached_hostname_info(self.cmc_name,
                                                    'control', 'hostname')

    async def get_status(self):
        """
        returns self.status
        either 'on' or 'off', or None if something wrong is going on
        """
        result = await self._get_cmc_verb('status')
        return result

    async def turn_on(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('on')
        return result

    async def turn_off(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('off')
        return result

    async def do_reset(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('reset')
        return result

    async def get_info(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('info', strip_result=False)
        return result

    async def get_usrpstatus(self):
        """
        returns self.usrpstatus
        either 'on' or 'off', or None if something wrong is going on
        """
        result = await self._get_cmc_verb('usrpstatus')
        return result

    async def turn_usrpon(self):
        """
        turn on node's USRP; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('usrpon')
        return result

    async def turn_usrpoff(self):
        """
        turn off node's USRP; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('usrpoff')
        return result

    async def turn_both_off(self):
        """
        turn off both USRP and node; expected result would be 'ok' if it goes fine
        """
        result1 = await self._get_cmc_verb('usrpoff')
        result2 = await self._get_cmc_verb('off')
        if result1 is None or result2 is None:
            return None
        return f"{result1} - {result2}"

    async def _get_cmc_verb(self, verb, strip_result=True):
        """
        verb typically is 'status', 'on', 'off' or 'info'
        """
        url = f"http://{self.cmc_name}/{verb}"
        cmc_timeout = float(Config().value('nodes', 'cmc_default_timeout'))
        http_timeout = aiohttp.ClientTimeout(connect=cmc_timeout, total=cmc_timeout)
        try:
            async with aiohttp.ClientSession(timeout=http_timeout) as session:
                async with session.get(url) as response:
                    text = await response.text(encoding='utf-8')
                    if strip_result:
                        text = text.strip()
                    setattr(self, verb, text)
        except aiohttp.client_exceptions.ClientConnectorError:
            logger.info(f"cannot connect to {url}")
            setattr(self, verb, None)
            return None
        except Exception:
            traceback.print_exc()
            setattr(self, verb, None)
            return None
        return getattr(self, verb)

    ####################
    # what status to expect after a message is sent
    expected_map = {
        'on': 'on',
        'reset': 'on',
        'off': 'off'
    }

    async def send_action(self, message="on", check=False, check_delay=1.):
        """
        Actually send action message like 'on', 'off' or 'reset'

        If check is True, waits for check_delay seconds
        before checking again that the status is what is expected, i.e.
        | message  | expected |
        |----------|----------|
        | on,reset | on       |
        | off      | off      |

        return value stored in self.action:

        * if check is false
          * True if request can be sent and returns 'ok',
            or None if something goes wrong
        * otherwise:
          * True to indicate that the node is correctly
            in 'on' mode after checking
          * False to indicate that the node is 'off' after checking
          * None if something goes wrong
        """
        url = f"http://{self.cmc_name}/{message}"
        cmc_timeout = float(Config().value('nodes', 'cmc_default_timeout'))
        http_timeout = aiohttp.ClientTimeout(connect=cmc_timeout, total=cmc_timeout)
        try:
            async with aiohttp.ClientSession(timeout=http_timeout) as session:
                async with session.get(url) as response:
                    text = await response.text(encoding='utf-8')
        except Exception:
            self.action = None
            return self

        is_ok = text.strip() == 'ok'

        if not check:
            self.action = is_ok
            return self
        await asyncio.sleep(check_delay)
        await self.get_status()
        self.action = self.status == self.expected_map[message]
        return self

    ####################
    message_to_reset_map = {'on': 'reset', 'off': 'on'}

    async def feedback(self, field, message):
        await self.message_bus.put(
            {'ip': self.control_ip_address(), field: message})

    async def ensure_reset(self):
        if self.status is None:
            await self.get_status()
        # still no status: means the CMC does not answer
        if self.status not in self.message_to_reset_map:
            await self.feedback(
                'reboot', f"Cannot get status at {self.cmc_name} (status={self.status})")
            return
        message_to_send = self.message_to_reset_map[self.status]
        await self.feedback(
            'reboot', f"Sending message '{message_to_send}' to CMC {self.cmc_name}")
        await self.send_action(message_to_send, check=True)
        if not self.action:
            await self.feedback(
                'reboot', f"Failed to send message {message_to_send} to CMC {self.cmc_name}")

    # used to be a coroutine, but since we need this
    # when dealing by KeybordInterrupt, it appears much safer
    # to just keep it a traditional function
    def manage_nextboot_symlink(self, action):
        """
        Messes with the symlink in /tftpboot/pxelinux.cfg/
        Depending on 'action'
        * 'cleanup' or 'harddrive' : clear the symlink
          corresponding to this CMC
        * 'frisbee' : define a symlink so that next boot
          will run the frisbee image
        see rhubarbe.conf for configurable options
        """

        the_config = Config()
        root = the_config.value('pxelinux', 'config_dir')
        frisbee = the_config.value('pxelinux', 'frisbee_image')

        # of the form 01-00-03-1d-0e-03-53
        mylink = "01-" + self.control_mac_address().replace(':', '-')
        source = os.path.join(root, mylink)

        if action in ('cleanup', 'harddrive'):
            if os.path.exists(source):
                logger.info(f"Removing {source}")
                os.remove(source)
        elif action in ('frisbee', ):
            if os.path.exists(source):
                os.remove(source)
            logger.info(f"Creating {source}")
            os.symlink(frisbee, source)
        else:
            logger.critical(
                f"manage_nextboot_symlink : unknown action {action}")


    ##########
    async def wait_for_telnet(self, service):
        ipaddr = self.control_ip_address()
        if service == 'frisbee':
            self.frisbee = Frisbee(ipaddr, self.message_bus)
            await self.frisbee.wait_until_connect()
        elif service == 'imagezip':
            self.imagezip = ImageZip(ipaddr, self.message_bus)
            await self.imagezip.wait_until_connect()

    async def reboot_on_frisbee(self, idle):
        self.manage_nextboot_symlink('frisbee')
        await self.ensure_reset()
        await self.feedback('reboot', f"idling for {idle}s")
        await asyncio.sleep(idle)

    async def run_frisbee(self, ipaddr, port, reset):
        await self.wait_for_telnet('frisbee')
        self.manage_nextboot_symlink('cleanup')
        result = await self.frisbee.run(ipaddr, port)
        #logger.info(f"run_frisbee -> {result}")
        if reset:
            await self.ensure_reset()
        else:
            await self.feedback('reboot',
                                'skipping final reset')
        return result

    async def run_imagezip(self, port, reset, radical, comment):
        await self.wait_for_telnet('imagezip')
        self.manage_nextboot_symlink('cleanup')
        result = await self.imagezip.run(port, self.control_hostname(),
                                         radical, comment)
        #logger.info(f"run_imagezip -> {result}")
        if reset:
            await self.ensure_reset()
        else:
            await self.feedback('reboot',
                                'skipping final reset')
        return result



"""
This implements a daemon that runs on faraday and that
takes care of the defined / authorized logins

think of it as a dedicated nodemanager
"""

# current implementation drawbacks:
# when migrating from an old box to a new box, it would make sense
# for this tool to enforce things that are metastable
# that is to say, moving stuff over from one box to the other
# can result in the homedir being present but without the .ssh dir, or
# with wrong ownership..
# This currently is not well taken care of, because all is done once when
# creating the account


# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w0703,w1202,w1202
# pylint: disable=logging-fstring-interpolation

import time
import os
import pwd
import logging
import subprocess
import shlex

from pathlib import Path

from rhubarbe.logger import accounts_logger as logger
from rhubarbe.config import Config
from rhubarbe.r2labapiproxy import R2labApiProxy

# accounts that the manager should leave alone
# xxx could be configurable
LEGIT_ACCOUNTS = {'faraday'}

def legal_name(name):
    return ('_' in name or '-' in name)

def run_command(command):
    """
    Run a shell command via subprocess.
    Returns True on success, False on failure (logged as error).
    """
    args = shlex.split(command)
    try:
        result = subprocess.run(
            args, capture_output=True, text=True, check=False)
        if result.returncode != 0:
            stderr = result.stderr.strip()
            logger.error(f"{command} -> rc={result.returncode}"
                         + (f": {stderr}" if stderr else ""))
            return False
        return True
    except OSError as exc:
        logger.error(f"{command} -> {exc}")
        return False

####################
# adapted from historical planetlab nodemanager

def replace_file_with_string(destination_path, new_contents,
                             owner=None, chmod=None,
                             remove_if_empty=False):
    """
    Replace a file with new contents
    checks for changes
      does not do anything if previous state was already right
    can handle chmod/chown if requested
    can also remove resulting file if contents are void, if requested

    returns
      * True if a change occurred, or the file is deleted
      * False if the file has no change
      * None if the file could not be created (e.g. directory missing)
    """
    try:
        with destination_path.open() as previous:
            current = previous.read()
    except IOError:
        current = ""
    if current == new_contents:
        # if turns out to be an empty string, and remove_if_empty is set,
        # then make sure to trash the file if it exists
        if remove_if_empty and not new_contents and destination_path.is_file():
            logger.info(
                f"replace_file_with_string: removing file {destination_path}")
            try:
                destination_path.unlink()
            except IOError as exc:
                logger.error(f"Could not remove file {destination_path}: {exc}")
            return True
        # we're done and have nothing to do
        return False
    # overwrite file: create a temp in the same directory
    try:
        with destination_path.open('w') as new:
            new.write(new_contents)
        if chmod:
            destination_path.chmod(chmod)
        if owner:
            run_command(f"chown {owner} {destination_path}")
        return True
    except IOError as exc:
        logger.error(f"Cannot create {destination_path}, {exc}")
        return None

####################


class AccountsManager:

    def __init__(self):
        the_config = Config()
        self.api_url = the_config.value('r2labapi', 'url')
        self.admin_token = the_config.value('r2labapi', 'admin_token')

        self._proxy = None

    def proxy(self):
        if self._proxy is None:
            self._proxy = R2labApiProxy(self.api_url, self.admin_token)
        return self._proxy

    @staticmethod
    def is_slice_account(record):
        """
        A slice account has uid >= 1001 and shell /bin/bash
        This filters out system accounts (systemd-network, etc.)
        and individual accounts with other shells
        explicity exclude the 'faraday' account that is used for other purposes
        """
        return (record.pw_name not in LEGIT_ACCOUNTS
                and legal_name(record.pw_name)
                and record.pw_uid >= 1001
                and record.pw_shell == '/bin/bash'
                and Path(record.pw_dir).exists())

    @classmethod
    def slices_from_passwd(cls):
        """
        Inspect /etc/passwd and return all slice logins
        """
        return [record.pw_name for record in pwd.getpwall()
                if cls.is_slice_account(record)]

    @staticmethod
    def slices_with_authorized():
        """
        Iterator:

        Inspect /home/ to find accounts that have authorized_keys.

        Focus on the ones that have a '_' in them, so that we leave
        alone custom accounts like 'guest' or similar.
        """
        homeroot = Path('/home')
        for authorized in homeroot.glob('*/.ssh/authorized_keys'):
            # need to move 2 steps up
            basename = authorized.parts[-3]
            if basename in LEGIT_ACCOUNTS:
                continue
            if not legal_name(basename):
                continue
            yield basename

    @staticmethod
    def create_account(slicename):
        """
        Does useradd with the right options
        Plus, creates .ssh dir with proper permissions
        Idempotent: skips steps already done, only logs real errors
        """
        homedir = Path(f"/home/{slicename}")
        sshdir = homedir / ".ssh"

        # create user only if it doesn't exist yet
        try:
            pwd.getpwnam(slicename)
        except KeyError:
            command = (f"useradd --create-home --user-group"
                       f" {slicename} --shell /bin/bash")
            logger.info(f"Running {command}")
            if not run_command(command):
                return

        # create .ssh dir if needed, ensure permissions
        if not sshdir.exists():
            logger.info(f"Creating {sshdir}")
            sshdir.mkdir(mode=0o700)
        else:
            sshdir.chmod(0o700)

        # ensure correct ownership (silent unless it fails)
        run_command(
            f"chown -R {slicename}:{slicename} {homedir}")


    # for #23
    @staticmethod
    def add_in_access_conf(slicename):
        # do the job of netsop-accessctl by hand
        # to get rid of the dependency
        lines = []
        with open("/etc/security/access.conf", encoding="utf-8") as f:
            for line in f:
                # nothing to do
                if line.startswith(f"+:{slicename}:"):
                    return
                # insert before the first "END local"
                if "END local" in line:
                    lines.append(f"+:{slicename}:ALL\n")
                lines.append(line)
        with open("/etc/security/access.conf", "w", encoding="utf-8") as f:
            f.writelines(lines)
        run_command("chmod 444 /etc/security/access.conf")


    @staticmethod
    def create_ssh_config(slicename):
        """
        Initialize slice's .ssh/config that keeps ssh from
        being too picky with host keys and similar

        Performed only if not yet existing
        """
        ssh_config_file = Path("/home") / slicename / ".ssh/config"

        # define the magic sequence for both fit* and data*
        config_bases = ['fit', 'data']
        #
        config_pattern = """Host {base}*
StrictHostKeyChecking no
UserKnownHostsFile=/dev/null
CheckHostIP=no
"""

        ssh_config = "\n".join(
            [config_pattern.format(base=base) for base in config_bases])
        replace_file_with_string(ssh_config_file,
                                 ssh_config,
                                 chmod=0o600,
                                 owner=f"{slicename}:{slicename}")

    @staticmethod
    def apply_keys(slicename, keys_string):
        auth_path = Path("/home") / slicename / ".ssh/authorized_keys"
        replace_file_with_string(auth_path,
                                 keys_string,
                                 chmod=0o600,
                                 owner=f"{slicename}:{slicename}",
                                 remove_if_empty=True)

    ##########
    def fetch_authorized_keys(self, slicename):
        """
        Fetch all SSH keys for all members of the given slice,
        using the /slices/by-name/{name}/keys endpoint.
        Returns a canonical authorized_keys string.
        """
        try:
            ssh_keys = self.proxy().get_slice_keys(slicename)
        except Exception as exc:
            logger.error(
                f"Could not fetch keys for slice {slicename}: {exc}")
            return ""
        key_lines = [(k['key'].replace("\n", "") + "\n")
                     for k in ssh_keys]
        key_lines.sort()
        return "".join(key_lines)

    ##########
    def manage_accounts(self, policy):             # pylint: disable=r0914

        try:
            slices = self.proxy().get_slices()
        except Exception as exc:
            logger.info(f"r2labapi unreachable ({exc}) - back to sleep")
            return

        current_slicenames = []
        if policy == 'leased':
            try:
                current_leases = self.proxy().get_current_leases()
                current_slicenames = [
                    lease['slice_name'] for lease in current_leases]
            except Exception as exc:
                logger.info(
                    f"r2labapi: cannot get leases ({exc}) - back to sleep")
                return

        # initialize with the slice names that are in /etc/passwd
        logins = self.slices_from_passwd()
        logger.debug(f"{len(logins)} current slice accounts in /etc/passwd:\n{logins}")

        # initialize map login_name -> authorized_keys contents
        # this is where we handle the fact that obsolete slices
        # will effectively have their authorized_keys voided
        auths_by_login = {login: "" for login in logins}

        for sliceobj in slices:
            slicename = sliceobj['name']
            # policy-dependant
            if policy == 'closed':
                authorized_keys = ""

            elif policy == 'leased':
                authorized_keys = ""
                if slicename in current_slicenames:
                    authorized_keys = self.fetch_authorized_keys(slicename)

            # policy == 'open'
            else:
                authorized_keys = self.fetch_authorized_keys(slicename)

            auths_by_login[slicename] = authorized_keys

        # implement it
        managed = 0
        for slicename, keys in auths_by_login.items():
            try:
                logger.debug(f"managing account {slicename} with {keys.count(chr(10))} keys")
                if slicename not in logins:
                    self.create_account(slicename)
                # do this always, allows to propagate later changes
                self.create_ssh_config(slicename)
                self.apply_keys(slicename, keys)
                # do this always too, allows to repair broken accounts
                self.add_in_access_conf(slicename)
                managed += 1
            except Exception:
                logger.exception(f"Could not deal with slice {slicename}")
        logger.debug(f"managed {managed} accounts")

    def run_forever(self, cycle, policy):
        while True:
            beg = time.time()
            logger.info("accounts manager "
                        f"policy = {policy}, cycle {cycle}s")
            self.manage_accounts(policy)
            now = time.time()
            duration = now - beg
            towait = cycle - duration
            if towait > 0:
                logger.info(f"----- accounts manager - "
                            f"sleeping for {towait:.2f}s")
                time.sleep(towait)
            else:
                logger.info(f"duration {duration}s exceeded cycle {cycle}s - "
                            f"skipping sleep")

    def main(self, cycle, debug):
        """
        cycle is the duration in seconds of one cycle

        Corner cases:
        * cycle = None : fetch value from config_bases
        * cycle = 0 : run just once (for debug mostly)
        """

        if debug:
            logger.setLevel(logging.DEBUG)
        if cycle is None:
            cycle = Config().value('accounts', 'cycle')
        cycle = int(cycle)
        policy = Config().value('accounts', 'access_policy')
        if policy not in ('open', 'leased', 'closed'):
            logger.error(f"Unknown policy {policy} - using 'closed'")
            policy = 'closed'
        # trick is
        if cycle != 0:
            self.run_forever(cycle, policy)
        else:
            logger.info(f"accounts manager oneshot "
                        f"policy = {policy}")
            self.manage_accounts(policy)



"""
The logic for loading an image on a set of nodes
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111

import asyncio

from asynciojobs import Scheduler, Job

from rhubarbe.frisbeed import Frisbeed
from rhubarbe.leases import Leases
from rhubarbe.config import Config


class ImageLoader:

    def __init__(self, nodes, image, bandwidth,         # pylint: disable=r0913
                 message_bus, display):
        self.nodes = nodes
        self.image = image
        self.bandwidth = bandwidth
        self.display = display
        self.message_bus = message_bus
        #
        self.frisbeed = None


    async def feedback(self, field, msg):
        await self.message_bus.put({field: msg})


    async def stage1(self):
        the_config = Config()
        idle = int(the_config.value('nodes', 'idle_after_reset'))
        await asyncio.gather(*[node.reboot_on_frisbee(idle)
                               for node in self.nodes])


    async def start_frisbeed(self):
        self.frisbeed = Frisbeed(self.image, self.bandwidth, self.message_bus)
        ip_port = await self.frisbeed.start()
        return ip_port


    async def stage2(self, reset):
        """
        wait for all nodes to be telnet-friendly
        then run frisbee in all of them
        and reset the nodes afterwards, unless told otherwise
        """
        # start_frisbeed will return the ip+port to use
        ipaddr, port = await self.start_frisbeed()
        results = await asyncio.gather(*[node.run_frisbee(ipaddr, port, reset)
                                         for node in self.nodes])
        # we can now kill the server
        self.frisbeed.stop_nowait()
        result = all(results)
        if not result:
            await self.feedback(
                'info',
                "at least one node failed to write that image on disk")
        return result


    # this is synchroneous
    def nextboot_cleanup(self):
        """
        Remove nextboot symlinks for all nodes in this selection
        so next boot will be off the harddrive
        """
        for node in self.nodes:
            node.manage_nextboot_symlink('harddrive')


    async def run(self, reset):
        leases = Leases(self.message_bus)
        await self.feedback('authorization', 'checking for a valid lease')
        valid = await leases.booked_now_by_me()
        if not valid:
            await self.feedback('authorization',
                                "Access refused : you have no lease "
                                "on the testbed at this time")
            return False
        await self.feedback('authorization', 'access granted')
        await (self.stage1()
               if reset
               else self.feedback('info', "Skipping stage1"))
        return await self.stage2(reset)


    def cleanup(self):
        if self.frisbeed:
            self.frisbeed.stop_nowait()
        self.nextboot_cleanup()
        self.display.epilogue()


    def main(self, reset, timeout):

        mainjob = Job(self.run(reset), critical=True)
        displayjob = Job(self.display.run(), forever=True, critical=True)
        scheduler = Scheduler(mainjob, displayjob,
                              timeout=timeout,
                              critical=False)

        try:
            is_ok = scheduler.run()
            if not is_ok:
                scheduler.debrief(silence_done_jobs=True)
                self.display.set_goodbye(
                    f"rhubarbe-load failed: {scheduler.why()}")
                return 1
            return 0 if mainjob.result() else 1
        except KeyboardInterrupt:
            self.display.set_goodbye(
                "rhubarbe-load : keyboard interrupt - exiting")
            return 1
        finally:
            self.cleanup()



#!/usr/bin/env python3

#pylint: disable=missing-module-docstring, missing-class-docstring, missing-function-docstring
import sys
import os
import traceback

import rhubarbe.main
from .selector import MisformedRange


class Rhubarbe:
    def main(self):
        """
        the entry point for all rhubarbe* commands
        """

        # use line buffering on stdout and stderr
        sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 1) # no buffering
        sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 1) # no buffering

        #
        supported = rhubarbe.main.supported_subcommands
        command = sys.argv[0]
        args = None
        subcommand = None
        # can we find a supported subcommand in argv[0]
        # i.e. when invoked as rhubarbe-monitornodes
        for candidate in supported:
            # just checking 'candidate in sys.argv[0]'
            # does not work as it would e.g. find subcommand
            # 'on' when trying to call monitornodes...
            # instead, we check that the command ends with e.g. -monitornodes
            if sys.argv[0].endswith("-" + candidate):
                subcommand = candidate
                args = sys.argv[1:]
                break

        # not found that way : subcommand must be argv[1]
        # so we need at least one argv left
        if not subcommand and len(sys.argv) <= 1:
            print(f"{command} needs a subcommand in {{{','.join(supported)}}}")
            exit(1)

        if not subcommand:
            subcommand = sys.argv[1]
            args = sys.argv[2:]
            if subcommand not in supported:
                print(f"Unknown subcommand {subcommand} "
                    f"- use one among {{{','.join(supported)}}}")
                exit(1)
        # do it
        entry_point = getattr(rhubarbe.main, subcommand)
        # remove subcommand from args
        try:
            exit(entry_point(*args))
        except MisformedRange as exc:
            print("ERROR: ", exc)
            exit(1)
        except Exception as exc:                        # pylint: disable=broad-except
            traceback.print_exc()
            print(f"{command} {subcommand} : Something went badly wrong : {exc}")
            exit(1)

def main():
    Rhubarbe().main()



# a fastapi based API to read relay temperatures
from fastapi import FastAPI, APIRouter
from pydantic import BaseModel
from typing import Optional

from ..inventoryrelays import InventoryRelays

inventory_relays = InventoryRelays.load()

app = FastAPI()

# define prefix /api/v1
api_v1 = APIRouter(prefix="/api/v1")


class RelayTemperatureQuery(BaseModel):
    duration: Optional[str] = None
    resample_period: Optional[str] = None


@api_v1.post("/relays/temperatures")
async def get_relays_temperatures(params: RelayTemperatureQuery):
    """
    Retrieve relay temperatures with optional filtering by duration and resample period.
    """
    # both parameters are optional and can be passed in a JSON body
    df = inventory_relays.load_past_data(
        duration=params.duration,
        resample_period=params.resample_period
    ).reset_index()
    return df.to_json(orient='records', date_format='iso')

app.include_router(api_v1)
def run_relays_api_service():
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=8000)



"""
Controller for the frisbee daemon that sends images when doing
rhubarbe load
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0201, r1705, w1201, w1202, w1203

from pathlib import Path

import asyncio

from rhubarbe.logger import logger
from rhubarbe.config import Config


class Frisbeed:
    """
    Controller for a frisbeed instance
    """
    def __init__(self, image, bandwidth, message_bus):
        self.image = str(image)
        self.bandwidth = bandwidth
        self.message_bus = message_bus
        #
        self.multicast_group = None
        self.multicast_port = None
        self.subprocess = None

    def __repr__(self):
        text = "<frisbeed"
        if self.multicast_group:
            text += f"@{self.multicast_group}:{self.multicast_port}"
        text += f" on {Path(self.image).name} at {self.bandwidth} Mibps"
        text += ">"
        return text

    async def feedback(self, field, msg):
        await self.message_bus.put({field: msg})

    def feedback_nowait(self, field, msg):
        self.message_bus.put_nowait({field: msg})

    async def start(self):                              # pylint: disable=r0914
        """
        Start a frisbeed instance
        returns a tuple multicast_group, port_number
        """
        the_config = Config()
        server = the_config.value('frisbee', 'server')
        server_options = the_config.value('frisbee', 'server_options')
        local_ip = the_config.local_control_ip()
        # in Mibps
        bandwidth = self.bandwidth * 2**20
        # should use default.ndz if not provided
        command_common = [
            server, "-i", local_ip, "-W", str(bandwidth), self.image
            ]
        # add configured extra options
        command_common += server_options.split()

        nb_attempts = int(the_config.value('networking', 'pattern_size'))
        pat_ip = the_config.value('networking', 'pattern_multicast')
        pat_port = the_config.value('networking', 'pattern_port')
        for i in range(1, nb_attempts+1):
            pat = str(i)
            multicast_group = pat_ip.replace('*', pat)
            multicast_port = str(eval(                  # pylint: disable=w0123
                pat_port.replace('*', pat)))
            command = command_common + [
                "-m", multicast_group, "-p", multicast_port,
                ]
            self.subprocess = await asyncio.create_subprocess_exec(
                *command,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.STDOUT
                )
            await asyncio.sleep(1)
            # after such a short time, frisbeed should not have returned yet
            # if it has, we try our luck on another couple (ip, port)
            command_line = " ".join(command)
            if self.subprocess.returncode is None:
                self.multicast_group = multicast_group
                self.multicast_port = multicast_port
                await self.feedback('info', f"started {self}")
                return multicast_group, multicast_port
            else:
                logger.warning(f"failed to start frisbeed with `{command_line}`"
                               f" -> {self.subprocess.returncode}")
        logger.critical(f"could not start frisbee server !!! on {self.image}")
        raise Exception(f"could not start frisbee server !!! on {self.image}")

    def stop_nowait(self):
        # make it idempotent
        if self.subprocess:
            self.subprocess.kill()
            self.subprocess = None
            self.feedback_nowait('info', f"stopped {self}")



"""
utility to send a given CMC action to a set of nodes
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w1202,r1705

from asynciojobs import Job, Scheduler

from rhubarbe.node import Node
from rhubarbe.display import Display


class Action:

    verb_to_method = {
        'status': 'get_status',
        'on': 'turn_on',
        'off': 'turn_off',
        'reset': 'do_reset',
        'info': 'get_info',
        'usrpstatus': 'get_usrpstatus',
        'usrpon': 'turn_usrpon',
        'usrpoff': 'turn_usrpoff',
        'bothoff': 'turn_both_off',
    }

    def __init__(self, verb, selector):
        self.verb = verb
        self.selector = selector

    async def get_and_show_verb(self, node, verb):
        assert verb in Action.verb_to_method
        # send the 'verb' method on node
        method = getattr(node, Action.verb_to_method[verb])
        # bound methods must not be passed the subject !
        result = await method()
        result = result if result is not None else f"{verb} N/A"
        for line in result.split("\n"):
            if line:
                print(f"{node.cmc_name}:{line}")

    # would make more sense to define this as a coroutine..
    def run(self, message_bus, timeout):
        """
        send verb to all nodes, waits for max timeout
        returns True if all nodes behaved as expected
        and False otherwise - including in case of KeyboardInterrupt
        """

        nodes = [Node(cmc_name, message_bus)
                 for cmc_name in self.selector.cmc_names()]
        jobs = [Job(self.get_and_show_verb(node, self.verb), critical=True)
                for node in nodes]
        display = Display(nodes, message_bus)
        scheduler = Scheduler(Job(display.run(), forever=True, critical=True),
                              *jobs,
                              timeout=timeout,
                              critical=False)
        try:
            if scheduler.run():
                return True
            else:
                scheduler.debrief(silence_done_jobs=True)
                print(f"rhubarbe-{self.verb} failed: {scheduler.why()}")
                return False
        except KeyboardInterrupt:
            print(f"rhubarbe-{self.verb} : keyboard interrupt - exiting")
            return False



"""
the class that model a relay box
the entry point for acquiring temperatures
"""

# pylint: disable=fixme, unspecified-encoding

from math import nan
from dataclasses import dataclass
from pathlib import Path
import asyncio

import pandas as pd

from importlib import resources
from dataclass_wizard import YAMLWizard
import asyncssh

from .config import Config
from .logger import logger

VERBOSE = False
# VERBOSE = True

# pylint: disable=missing-function-docstring, missing-class-docstring
def verbose(*args, **kwds):
    if not VERBOSE:
        return
    print(*args, **kwds)



@dataclass
class Relay:

    host: str
    # IP: str                                     # pylint: disable=invalid-name

    async def get_temperature(self):
        async with asyncssh.connect(self.host) as conn:
            result = await conn.run('/usr/bin/vcgencmd measure_temp', check=True)
        raw1 = result.stdout
        raw2 = raw1.split('=')[1]
        raw3 = raw2.split("'")[0]
        return float(raw3)


    def store_current_temperature(self, temperature):
        now = (
            pd.Timestamp.utcnow()
            .replace(microsecond=0)
            .isoformat()
            .replace("+00:00","Z")
        )
        # temperature = asyncio.run(self.get_temperature())
        folder = Path(Config().value('testbed', 'relays_database_folder'))
        if not folder.is_dir():
            print(f"Creating folder {folder}")
            folder.mkdir(parents=True, exist_ok=True)
        with (folder / "temperatures.csv").open('a') as writer:
            print(f"{now},{self.host},{temperature}", file=writer)


@dataclass
class InventoryRelays(YAMLWizard):

    relays: list[Relay]

    @staticmethod
    def load() -> "InventoryRelays":
        the_config = Config()
        yaml_path = the_config.value('testbed', 'inventory_relays_path')
        try:
            with open(yaml_path) as feed:
                return InventoryRelays.from_yaml(feed.read())
        except FileNotFoundError:
            # not all deployments have relays
            logger.warning(f"file not found {yaml_path}")
            return InventoryRelays([])
        except KeyError as exc:
            print(f"something wrong in config file {yaml_path}, {exc}")
            raise


    def get_temperatures(self, *, mode):
        """
        mode is expected to be either 'print' or 'save'
        """
        async def run_all():
            return await asyncio.gather(
                *(relay.get_temperature() for relay in self.relays)
            )

        temperatures = asyncio.run(run_all())
        match mode:
            case 'print':
                for relay, temperature in zip(self.relays, temperatures):
                    print(f"{relay} has temperature {temperature:.2f}C")
            case 'store':
                for relay, temperature in zip(self.relays, temperatures):
                    relay.store_current_temperature(temperature)

    def __iter__(self):
        return iter(self.relays)
    def __len__(self):
        return len(self.relays)

    def load_past_data(self, *, duration=None, resample_period=None):
        folder = Path(Config().value('testbed', 'relays_database_folder'))
        df = pd.read_csv(
            folder / "temperatures.csv",
            names=["timestamp", "relay", "temperature"])
        df['timestamp'] = pd.to_datetime(df['timestamp'], format="ISO8601")
        df.set_index('timestamp', inplace=True)
        if duration is not None:
            duration = pd.Timedelta(duration)
            time_threshold = pd.to_datetime(pd.Timestamp.utcnow()) - duration
            df = df[df.index >= time_threshold]
        if resample_period is not None:
            # resample already has string builtin conversion
            df = df.groupby('relay').resample(resample_period).mean()
        return df



from dataclasses import dataclass
from dataclass_wizard import YAMLWizard

@dataclass
class KO(YAMLWizard):
    items: list[str]

    def list(self):
        pass

@dataclass
class OK(YAMLWizard):
    items: list[str]



"""
Make a class a singleton
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202


class Singleton(type):
    """
    a singleton metaclass
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = \
                super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# trap signals for a more condensed output
import signal
import functools

import asyncio

from rhubarbe.logger import monitor_logger as logger


class MonitorLoop:

    def __init__(self, message):
        self.message = message

    def run(self, async_main):

        async def async_main_wrapper():
            loop = asyncio.get_running_loop()

            def exiting(signame):
                logger.info(f"rhubarbe-{self.message}: ¬ received signal {signame} - exiting")
                loop.stop()
                exit(1)

            for signame in ('SIGHUP', 'SIGQUIT', 'SIGINT', 'SIGTERM'):
                loop.add_signal_handler(getattr(signal, signame),
                                        functools.partial(exiting, signame))

            try:
                await async_main
                return 0
            except KeyboardInterrupt:
                logger.info(f"rhubarbe-{self.message} : keyboard interrupt - exiting")
                return 1
            except asyncio.TimeoutError:
                logger.info(f"rhubarbe-{self.message} : asyncio timeout expired")
                return 1

        with asyncio.Runner() as runner:
            return runner.run(async_main_wrapper())



"""
REST client for the r2labapi service.
Replaces the XMLRPC-based PlcApiProxy.
"""

# c0111 no docstrings yet
# w1202 logger & format
# pylint: disable=c0111, w1202

import os
import getpass
from datetime import datetime, timezone

import requests

from .logger import logger


def iso_to_epoch(iso_string):
    """Convert ISO 8601 datetime string to Unix epoch."""
    # handle Z suffix for Python 3.10 compat
    if iso_string.endswith('Z'):
        iso_string = iso_string[:-1] + '+00:00'
    return datetime.fromisoformat(iso_string).timestamp()


def epoch_to_iso(epoch):
    """Convert Unix epoch to ISO 8601 UTC datetime string."""
    return datetime.fromtimestamp(epoch, tz=timezone.utc).isoformat()


class R2labApiProxy:
    """
    A thin REST client for the r2labapi service.

    Two authentication modes:
    - pre-shared admin token: R2labApiProxy(url, admin_token="...")
    - interactive login: R2labApiProxy(url) then write operations
      will prompt for email/password (or read R2LABAPI_EMAIL /
      R2LABAPI_PASSWORD env vars)
    """

    def __init__(self, url, admin_token=None):
        self.url = url.rstrip('/')
        self.session = requests.Session()
        self.session.headers['Accept'] = 'application/json'
        if admin_token:
            self.session.headers['Authorization'] = \
                f'Bearer {admin_token}'

    @property
    def is_authenticated(self):
        return 'Authorization' in self.session.headers

    def login(self, email, password):
        """Authenticate with email/password, store JWT."""
        response = self.session.post(
            f'{self.url}/auth/login',
            json={'email': email, 'password': password})
        response.raise_for_status()
        token = response.json()['access_token']
        self.session.headers['Authorization'] = f'Bearer {token}'

    def ensure_authenticated(self):
        """Prompt for credentials if not already authenticated."""
        if self.is_authenticated:
            return
        email = (os.environ.get("R2LABAPI_EMAIL")
                 or input("Enter r2labapi email (login): "))
        password = (os.environ.get("R2LABAPI_PASSWORD")
                    or getpass.getpass(
                        f"Enter r2labapi password for {email}: "))
        self.login(email, password)

    # ---- low-level HTTP verbs ----

    def _get(self, path, params=None):
        url = f'{self.url}{path}'
        response = self.session.get(url, params=params)
        response.raise_for_status()
        return response.json()

    def _post(self, path, json=None):
        self.ensure_authenticated()
        url = f'{self.url}{path}'
        response = self.session.post(url, json=json)
        response.raise_for_status()
        if response.status_code == 204:
            return None
        return response.json()

    def _patch(self, path, json=None):
        self.ensure_authenticated()
        url = f'{self.url}{path}'
        response = self.session.patch(url, json=json)
        response.raise_for_status()
        return response.json()

    def _delete(self, path):
        self.ensure_authenticated()
        url = f'{self.url}{path}'
        response = self.session.delete(url)
        response.raise_for_status()

    # ---- slices ----

    def get_slices(self):
        """
        GET /slices
        Returns list of SliceRead:
          {id, name, family, member_ids, ...}
        """
        return self._get('/slices')

    def get_slice_keys(self, slicename):
        """
        GET /slices/by-name/{name}/keys
        Returns list of SSHKeyRead:
          {id, key, comment, created_at}
        """
        return self._get(f'/slices/by-name/{slicename}/keys')

    # ---- leases ----

    def get_leases(self, **params):
        """
        GET /leases
        Optional filters: resource_id, slice_id, alive (epoch),
                          after (ISO), before (ISO)
        Returns list of LeaseRead:
          {id, resource_id, slice_id, t_from, t_until, slice_name}
        """
        return self._get('/leases', params=params or None)

    def get_current_leases(self):
        """Convenience: get leases alive right now."""
        import time
        now = int(time.time())
        return self._get('/leases', params={'alive': now})

    def create_lease(self, body):
        """
        POST /leases
        body: {resource_name, slice_name, t_from (ISO), t_until (ISO)}
        Returns LeaseRead
        """
        return self._post('/leases', json=body)

    def update_lease(self, lease_id, body):
        """
        PATCH /leases/{lease_id}
        body: {t_from? (ISO), t_until? (ISO)}
        Returns LeaseRead
        """
        return self._patch(f'/leases/{lease_id}', json=body)

    def delete_lease(self, lease_id):
        """DELETE /leases/{lease_id}"""
        self._delete(f'/leases/{lease_id}')

    # ---- resources ----

    def get_resource_by_name(self, name):
        """
        GET /resources/by-name/{name}
        Returns ResourceRead: {id, name, granularity}
        """
        return self._get(f'/resources/by-name/{name}')

    def __str__(self):
        return f"R2labApiProxy@{self.url}"



"""
The pdus monitor cyclically checks for the status of all pdus,
and reports it to the sidecar service

As a start, the tool only reports ON or OFF or UNKNOWN
"""

# pylint: disable=logging-fstring-interpolation, fixme, missing-function-docstring

import asyncio
from math import nan

from rhubarbe.logger import monitor_logger as logger

from rhubarbe.inventorypdus import InventoryPdus, PduDevice
from rhubarbe.monitor.reconnectable import ReconnectableSidecar


class MonitorPdu:

    """
    monitor one PDU
    """

    def __init__(self, pdu_device: PduDevice,
                 reconnectable, verbose, cycle=2):
        self.pdu_device = pdu_device
        self.reconnectable = reconnectable
        self.cycle = cycle
        self.verbose = verbose
        self.info = {'id': self.name}

    @property
    def name(self):
        return self.pdu_device.name

    def __repr__(self):
        return f"monitored pdu #{self.name}"

    async def emit(self):
        await self.reconnectable.emit_info(self.info)

    async def probe(self):
        # avoid clogging the logs
        status = await self.pdu_device.status(show_stdout=False)
        on_off = 'on' if status == 0 else 'off' if status == 1 else 'unknown'
        self.info['on_off'] = on_off
        self.publish_extras()
        if self.verbose:
            logger.info(f"on_off on PDU {self.name} is {on_off}")
        await self.emit()

    def publish_extras(self):
        """
        propagate positions information and other cosmetic information
        """
        dev = self.pdu_device
        if dev.icon_x_rank is nan and dev.icon_y_rank is nan:
            pass
        elif dev.icon_x_rank is not nan and dev.icon_y_rank is not nan:
            logger.warning(
                f"pdu {dev.name} has both icon_x_rank and icon_y_rank defined - ignored"
            )
        elif dev.icon_x_rank is not nan:
            self.info['icon_x_rank'] = dev.icon_x_rank
            self.info['icon_y_rank'] = 0
            self.info['icon_units'] = dev.icon_units
        else:
            self.info['icon_x_rank'] = 0
            self.info['icon_y_rank'] = dev.icon_y_rank
            self.info['icon_units'] = dev.icon_units
        if dev.location_x_grid is not nan:
            self.info['location_x_grid'] = dev.location_x_grid
        if dev.location_y_grid is not nan:
            self.info['location_y_grid'] = dev.location_y_grid
        if dev.label:
            self.info['label'] = dev.label

    async def probe_forever(self):
        while True:
            await self.probe()
            await asyncio.sleep(self.cycle)


class MonitorPdus:
    """
    monitor all phones status and report to the sidecar service
    """

    def __init__(self, verbose, sidecar_url, cycle, names=None):
        self.verbose = verbose

        devices = InventoryPdus.load().devices
        if names:
            devices = [device for device in devices if device.name in names]

        self.reconnectable = ReconnectableSidecar(sidecar_url, 'pdus')
        # xxx this is fragile
        # we rely on the fact that the items in the inventory
        # match the args of MonitorPdu's constructor
        self.pdus = [
                MonitorPdu(
                    pdu_device = device,
                    reconnectable=self.reconnectable,
                    verbose=verbose,
                    cycle=cycle)
                for device in devices
            ]

    async def run_forever(self):
        await asyncio.gather(
            *[pdu.probe_forever()
              for pdu in self.pdus],
            self.reconnectable.keep_connected())



"""
the classes that help manage our PDU devices
"""

# pylint: disable=fixme, unspecified-encoding

from math import nan
from dataclasses import dataclass
import asyncio

from importlib import resources
from dataclass_wizard import YAMLWizard
import asyncssh

from .config import Config
from .logger import logger

VERBOSE = False
#VERBOSE = True

# pylint: disable=missing-function-docstring, missing-class-docstring
def verbose(*args, **kwds):
    if not VERBOSE:
        return
    print(*args, **kwds)



@dataclass
class PduHost:

    name: str
    type: str
    IP: str                                     # pylint: disable=invalid-name
    username: str = ""
    password: str = ""
    chain_length: int = 1


    def oneline(self):
        text = ""
        if self.username == "":
            text += f"no ssh"
        else:
            text += f"ssh = {self.username}@{self.IP}"
        if self.is_chained():
            text += f" / daisy chain of {self.chain_length} boxes"
        else:
            text += " / no chaining"
        return text


    def is_chained(self):
        return self.chain_length > 1


    async def run_pdu_shell(self, action, device_name, *args, show_stdout=True):
        """
        run the 'pdu' command on the PDU host where this input is attached
        when this targets a device, the device_name is set - otherwise it is None
        """
        env = dict(PDU_USERNAME=self.username, PDU_PASSWORD=self.password)
        # contains sensitive information
        # verbose(10*'-', "debug: the script configuration:")
        # for key, value in env.items():
        #     verbose(f'export {key}="{value}"')
        # verbose(10*'-')
        script = f"scripts/{self.type}"
        command_path = (resources.files("rhubarbe") / script)
        if not command_path.exists():
            print(f"Could not locate script '{script}' - exiting")
            return 255

        command = f"{command_path} {action} {self.IP} {device_name or '""'} {' '.join(str(arg) for arg in args)}"
        verbose(f"PduHost: running command '{command}' with {device_name=}")
        proc = await asyncio.create_subprocess_shell(
            command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env)
        stdout, stderr = await proc.communicate()
        verbose(f"{proc.returncode=}")
        if proc.returncode == 255:
            print("FAILURE:", stdout.decode(), end="")
            return 255
        verbose("SUCCESS")
        if stdout and show_stdout:
            print(stdout.decode(), end="")
        if stderr:
            print("STDERR", stderr.decode(), end="")
        return proc.returncode


    async def probe(self):
        return await self.run_pdu_shell("probe", None)


@dataclass
class PduInput:

    pdu_host_name: str
    outlet: int
    in_chain: int = 0
    # not in the YAML, will be located
    # after loading
    pdu_host: PduHost = None


    def __repr__(self):
        return f"{self.pdu_host_name}:{self.oneline()}"


    def oneline(self):
        if self.in_chain == 0:
            return f"        outlet {self.outlet}"
        return f"chain-{self.in_chain}@outlet-{self.outlet}"

    async def run_pdu_shell(self, action, *args, device_name="", show_stdout=True):
        """
        run the 'pdu' command on the PDU host where this input is attached
        """
        if device_name:
            message = f"PduInput: running action '{action}' on device {device_name}"
        verbose(message)
        return await self.pdu_host.run_pdu_shell(action, device_name, *args, show_stdout=show_stdout)




@dataclass
class PduDevice:
    name: str
    inputs: list[PduInput]
    description: str = ""
    ssh_hostname: str = ""
    ssh_username: str = "root"
    # will be maintained by actions made
    status_cache: bool | None = None
    # if set to True, the device will be turned off when the testbed is idle
    auto_turn_off: bool = False
    icon_x_rank: float = nan
    icon_y_rank: float = nan
    icon_units: str = "grid"  # can be set to 'grid' or 'rank'
    location_x_grid: float = nan
    location_y_grid: float = nan
    label: str = ""


    async def is_pingable(self, timeout=1) -> bool:
        """
        when there is a ssh_hostname attached, does that answer ping ?
        """
        if self.ssh_hostname is None:
            return False
        command = f"ping -c 1 -w {timeout} {self.ssh_hostname}"
        proc = await asyncio.create_subprocess_shell(
            command,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL)
        # failure occurs through timeout
        returncode = await asyncio.wait_for(proc.wait(), timeout=2)
        return returncode == 0


    async def turn_off_through_ssh_if_pingable(self) -> bool:
        """
        if the device is pingable, turn it off through ssh

        returns True if device is pingable and the ssh command succeeds
        """
        is_pingable = await self.is_pingable()
        if not is_pingable:
            return False
        print(f"Doing a soft TURN OFF on device {self.name}")
        command = "shutdown -h now"
        ssh_options = {
            'known_hosts': None,
            'username': "root",
        }

        async with asyncssh.connect(self.ssh_hostname, **ssh_options) as conn:
            completed = await conn.run(command)
            return completed.returncode == 0


    async def attempt_soft_reset(self):
        soft_turned_off = await self.turn_off_through_ssh_if_pingable()
        # give it some time to complete the soft OFF
        # xxx could be configurable
        if soft_turned_off:
            await asyncio.sleep(15)


    async def run_pdu_shell_on_all_inputs(self, action, *, show_stdout=True):
        retcod = await asyncio.gather(
            *(input.run_pdu_shell(
                action, input.in_chain, input.outlet, device_name = self.name,
                show_stdout=show_stdout)
            for input in self.inputs)
        )
        return retcod


    async def _status_or_on(self, action, *, show_stdout=True):
        """
        the ON and STATUS actions are similar in their logic
        - we are sure that the node is ON if any input is ON
        - we are sure that the node is OFF if all inputs are OFF
        - otherwise, we don't know for sure
        """
        retcods = await self.run_pdu_shell_on_all_inputs(
            action, show_stdout=show_stdout)
        # if all inputs fail, we can't say
        if all(retcod == 255 for retcod in retcods):
            self.status_cache = 255
        # if any input is ON then the device is ON
        elif any(retcod == 0 for retcod in retcods):
            self.status_cache = 0
        # if any input is unknown, then we can't say
        elif any(retcod == 255 for retcod in retcods):
            self.status_cache = 255
        # else we are sure the node is OFF
        else:
            self.status_cache = 1
        verbose(f"{self.name} status_cache is now {self.status_cache}")
        return self.status_cache


    async def on(self):                                 # pylint: disable=invalid-name
        return await self._status_or_on('on')

    async def status(self, show_stdout=True):
        return await self._status_or_on('status', show_stdout=show_stdout)

    async def off(self):
        """
        the OFF action
        start with attempting a soft reset
        after that, the logic is simple because
        * retcod can only be 0 or 255 (not 1)
        * and retcod == 0 means the OFF has succeeded
        """
        await self.attempt_soft_reset()
        retcods = await self.run_pdu_shell_on_all_inputs('off')
        # if all inputs say 0 (they were turned off), node is off
        if all(retcod == 0 for retcod in retcods):
            self.status_cache = 1
        else:
            self.status_cache = 255
        return self.status_cache


    async def reset(self):
        """
        the RESET action
        """
        off = await self.off()
        if off != 1:
            return 255
        # xxx could be configurable
        await asyncio.sleep(15)
        return await self.on()



    async def run_action(self, action):
        """
        returns the main retcod as defined above (0=OK/ON 1=OK/OFF 255=KO)
        also update self.status when relevant

        when a device has several inputs:

        STATUS returns
            0 if at least one input is ON
            1 if all inputs are OFF
            255 otherwise

        ON will try to turn on all inputs
           if AT LEAST ONE works fine, the return code is 0 - 255 otherwise

        OFF will try to turn off all inputs
           if ALL work fine, the return code is 1 - 255 otherwise

        RESET mainly does
            OFF; if fails, reset fails
            sleep for a bit, ON, and returns what ON returns

        OFF and RESET will first try to do a "soft" turn-off using ssh
        (if contigured)
        they go on regardless to do the "hard" turn-off, after some delay
        """

        match action:
            case 'on':
                return await self.on()
            case 'off':
                return await self.off()
            case 'status':
                return await self.status()
            case 'reset':
                return await self.reset()
            case _:
                print(f"OOPS: unknown action '{action}'")
                return 255



@dataclass
class InventoryPdus(YAMLWizard):

    # see below the `display_list` method and why it was renamed from 'list'
    pdu_hosts: list[PduHost]
    devices: list[PduDevice]


    @staticmethod
    def load() -> "InventoryPdus":
        the_config = Config()
        yaml_path = the_config.value('testbed', 'inventory_pdus_path')
        try:
            with open(yaml_path) as feed:
                return InventoryPdus.from_yaml(feed.read()).solve_references()
        except FileNotFoundError:
            # not all deployments have pdus
            logger.warning(f"file not found {yaml_path}")
            return InventoryPdus([], [])
        except KeyError as exc:
            print(f"something wrong in config file {yaml_path}, {exc}")
            raise


    def solve_references(self):
        """
        fill all PduInput instances with their pdu_host attribute
        """
        hosts_by_name = {pdu_host.name: pdu_host for pdu_host in self.pdu_hosts}
        for device in self.devices:
            for input_ in device.inputs:
                input_.pdu_host = hosts_by_name[input_.pdu_host_name]
        return self


    def status(self):
        """
        displays the status for all known devices
        works sequentially on all hosts so that the output is readable
        """
        print(f"we have {len(self.pdu_hosts)} PDUs and {len(self.devices)} devices. ")
        pdu_host_width = max(len(pdu_host.name) for pdu_host in self.pdu_hosts)
        type_width = max(len(pdu_host.type) for pdu_host in self.pdu_hosts)
        sep = 10 * '='

        async def status_all():
            for pdu_host in self.pdu_hosts:
                print(f"{sep} {pdu_host.name:>{pdu_host_width}} ({pdu_host.type:<{type_width}}) {sep}")
                print(f"{pdu_host.oneline()}")

                await pdu_host.probe()
        with asyncio.Runner() as runner:
            runner.run(status_all())


    # this method used to be called plain 'list'
    # however when moving to 3.14 it started to conflict with builtin 'list'
    # especially in the definition of the dataclass fields
    # and that was screwing up dataclass_wizard; not quite sure how that happened
    # see also https://github.com/rnag/dataclass-wizard/issues/219
    def display_list(self, names=None):
        """
        if no name: list all pdu hosts
        otherwise, list all pdu_hosts AND all pdu_devices
        whose name is in the list (case ignored)
        """
        if not names:
            print(f"we have {len(self.pdu_hosts)} PDUs and {len(self.devices)} devices. "
                  f"(*) means auto_turn_off")

        names = [] if names is None else [n.lower() for n in names]
        pdu_host_width = max(len(pdu_host.name) for pdu_host in self.pdu_hosts)
        type_width = max(len(pdu_host.type) for pdu_host in self.pdu_hosts)
        device_width = max(len(device.name) for device in self.devices)
        sep = 10 * '='
        indent_empty = 5 * ' '
        indent_auto = ' (*) '
        for pdu_host in self.pdu_hosts:
            # if no name was passed, list all pdu_hosts
            if names and pdu_host.name.lower() not in names:
                continue
            print(f"{sep} {pdu_host.name:>{pdu_host_width}} ({pdu_host.type:<{type_width}}) {sep}")
            print(f"{pdu_host.oneline()}")
            for device in self.devices:
                for input_ in device.inputs:
                    indent = indent_auto if device.auto_turn_off else indent_empty
                    if input_.pdu_host_name == pdu_host.name:
                        print(f"{indent}{input_.oneline()} "
                              f"→ {device.name:<{device_width}}")

        # if no name was passed, stop here
        if not names:
            return
        for device in self.devices:
            if device.name.lower() not in names:
                continue
            print(f"{sep} device {device.name:^{device_width}} {sep}")
            for input_ in device.inputs:
                indent = indent_auto if device.auto_turn_off else indent_empty
                print(f"{indent}{input_}")

    def _get_object(self, name, attribute, kind):
        l_objs = getattr(self, attribute)
        for obj in l_objs:
            if obj.name == name:
                return obj
        raise ValueError(f"unknown {kind} '{name}'")
    def get_device(self, name) -> PduDevice:
        return self._get_object(name, 'devices', 'device')
    def get_pdu_host(self, name) -> PduHost:
        return self._get_object(name, 'pdu_hosts', 'pdu_host')



"""
The nodes monitor cyclically checks for the status of all nodes,
and reports it to the sidecar service
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return

# pylint: disable=fixme,logging-fstring-interpolation,missing-function-docstring

import re
import asyncio

from rhubarbe.config import Config
from rhubarbe.node import Node
from rhubarbe.ssh import SshProxy
# use a dedicated logger for monitors
from rhubarbe.logger import monitor_logger as logger

# connect to sidecar
from rhubarbe.monitor.reconnectable import ReconnectableSidecar

# translate info into a single char for logging
def one_char_summary(info):
    if 'cmc_on_off' in info and info['cmc_on_off'] != 'on':
        return '.'
    if 'control_ping' in info and info['control_ping'] != 'on':
        return 'o'
    if 'control_ssh' in info and info['control_ssh'] != 'on':
        return '0'
    if 'os_release' in info and 'fedora' in info['os_release']:
        return 'F'
    if 'os_release' in info and 'ubuntu' in info['os_release']:
        return 'U'
    return '^'


class MonitorNode:
    """
    the logic for probing a node as part of monitoring
    formerly in fitsophia/website/monitor.py
    . first probes for the CMC status; if off: done
    . second probe for ssh; if on : done
    . third, checks for ping
    """

    def __init__(self, node, reconnectable,             # pylint: disable=r0913
                 verbose=False):
        # a rhubarbe.node.Node instance
        self.node = node
        self.reconnectable = reconnectable
        self.verbose = verbose
        # current info - will be reported to sidecar
        self.info = {'id': node.id}

    def set_info(self, *overrides):
        """
        update self.info with all the dicts in overrides
        """
        for override in overrides:
            self.info.update(override)

    async def report_info(self):
        """
        Send info to sidecar
        """
        await self.reconnectable.emit_info(self.info)

    async def set_info_and_report(self, *overrides):
        """
        Convenience to set info and immediately report
        """
        self.set_info(*overrides)
        await self.report_info()


    ubuntu_matcher = re.compile(r"DISTRIB_RELEASE=(?P<ubuntu_version>[0-9.]+)")
    fedora_matcher = re.compile(r"Fedora release (?P<fedora_version>\d+)")
    centos_matcher = re.compile(r"CentOS Linux release (?P<centos_version>\d+)")
    gnuradio_matcher = re.compile(
        r"\AGNURADIO:(?P<gnuradio_version>[0-9\.]+)\Z")
    uname_matcher = re.compile(r"\AUNAME:(?P<uname>.+)\Z")
    # 2016-05-28@08:20 - node fit38 - image oai-enb-base2 - by root
    rhubarbe_image_matcher = re.compile(
        r"\A/etc/rhubarbe-image:.* - image (?P<image_radical>[^ ]+) - by")
    docker_matcher = re.compile(
        r"\ADOCKER:Docker version (?P<docker_version>[0-9\.]+),")
    container_matcher = re.compile(
        r"\ACONTAINER:(?P<running>(true|false)) (?P<image>.*)")


    def parse_ssh_probe_output(self,      # pylint: disable=r0912, r0914, r0915
                               stdout, padding_dict):
        os_release = "other"
        gnuradio_release = "none"
        uname = ""
        image_radical = ""
        docker_version = ""
        container_running = ""
        container_image = ""
        for line in stdout.split("\n"):
            match = self.ubuntu_matcher.match(line)
            if match:
                version = match.group('ubuntu_version')
                os_release = f"ubuntu-{version}"
                continue
            match = self.fedora_matcher.match(line)
            if match:
                version = match.group('fedora_version')
                os_release = f"fedora-{version}"
                continue
            match = self.centos_matcher.match(line)
            if match:
                version = match.group('centos_version')
                os_release = f"centos-{version}"
                continue
            match = self.gnuradio_matcher.match(line)
            if match:
                gnuradio_release = match.group('gnuradio_version')
                continue
            match = self.uname_matcher.match(line)
            if match:
                uname = match.group('uname')
                continue
            match = self.rhubarbe_image_matcher.match(line)
            if match:
                image_radical = match.group('image_radical')
                continue
            if match := self.docker_matcher.match(line):
                docker_version = match.group('docker_version')
                continue
            if match := self.container_matcher.match(line):
                container_running = match.group('running')
                container_image = match.group('image')
                continue

        # xxx would make sense to clean up history for measurements that
        # we were not able to collect at this cycle
        self.set_info({
            'os_release': os_release,
            'gnuradio_release': gnuradio_release,
            'uname': uname,
            'image_radical': image_radical,
            'docker_version': docker_version,
            'container_running': container_running,
            'container_image': container_image,
            }, padding_dict)

    async def probe(self,                 # pylint: disable=r0912, r0914, r0915
                    ping_timeout, ssh_timeout):
        """
        The logic for getting one node's info and send it to sidecar
        """
        if self.verbose:
            logger.info(f"entering pass1, info={self.info}")
        # pass1 : check for status
        padding_dict = {
            'control_ping': 'off',
            'control_ssh': 'off',
            # don't overwrite os_release though
        }
        # get USRP status no matter what - use "" if we receive None
        # to limit noise when the node is physically removed
        usrp_status = await self.node.get_usrpstatus() or 'fail'
        # replace usrpon and usrpoff with just on and off
        self.set_info({'usrp_on_off': usrp_status.replace('usrp', '')})
        # get CMC status
        status = await self.node.get_status()
        if status == "off":
            await self.set_info_and_report({'cmc_on_off': 'off'}, padding_dict)
            return
        elif status != "on":
            await self.set_info_and_report({'cmc_on_off': 'fail'}, padding_dict)
            return
        if self.verbose:
            logger.info(f"entering pass2, info={self.info}")
        # pass2 : CMC status is ON - let's try to ssh it
        self.set_info({'cmc_on_off': 'on'})
        padding_dict = {
            'control_ping': 'on',
            'control_ssh': 'on',
        }

        remote_commands = [
            "cat /etc/lsb-release /etc/redhat-release /etc/gnuradio-release "
            "2> /dev/null | grep -i release",
            "echo -n GNURADIO: ; gnuradio-config-info --version "
            "2> /dev/null || echo none",
            # this trick allows to have the filename on each output line
            "grep . /etc/rhubarbe-image /dev/null",
            "echo -n UNAME: ; uname -r",
            "echo -n DOCKER: ; docker --version",
            "echo -n CONTAINER: ; docker inspect "
                "--format='{{.State.Running}} {{.Config.Image}}' container",
            ]
        # reconnect each time
        self.set_info({'control_ssh': 'off'})
        async with SshProxy(self.node) as ssh:
            logger.info(f"trying to ssh-connect to {self.node.control_hostname()} "
                        f"(timeout={ssh_timeout})")
            try:
                connected = await asyncio.wait_for(
                    ssh.connect(), timeout=ssh_timeout)
            except asyncio.TimeoutError:
                connected = False
                self.set_info({'control_ssh': 'off'})
            logger.info(f"{self.node.control_hostname()} ssh-connected={connected}")
            if connected:
                self.set_info({'control_ssh': 'on'})
                try:
                    command = ";".join(remote_commands)
                    output = await asyncio.wait_for(
                        ssh.run(command), timeout=ssh_timeout)
                    logger.debug(f"{output=:20s}...")
                    # padding dict here sets control_ssh and control_ping to on
                    self.parse_ssh_probe_output(output, padding_dict)
                    # required as otherwise we leak openfiles
                    try:
                        await ssh.close()
                    except Exception:                           # pylint: disable=broad-except
                        logger.exception("monitornodes oops 1")
                except asyncio.TimeoutError:
                    logger.info(f"received ssh timeout with {self.node.control_hostname()}")
                    self.set_info({'control_ssh': 'off'})
                except Exception:                                   # pylint: disable=broad-except
                    logger.exception("monitornodes remote_command failed")
        logger.info(f"{self.node.control_hostname()} ssh-based logic done "
                    f"ssh is deemed {self.info['control_ssh']}")

        # if we could ssh then we're done
        if self.info['control_ssh'] == 'on':
            await self.report_info()
            return

        logger.info(f"entering pass3, info={self.info}")
        # pass3 : node is ON but could not ssh
        # check for ping
        # I don't know of an asyncio library to deal with icmp
        # so let's use asyncio.subprocess
        # xxx maybe a Ping class would be the way to go
        control = self.node.control_hostname()
        command = ["ping", "-c", "1", "-t", "1", control]
        try:
            subprocess = await asyncio.create_subprocess_exec(
                *command,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL)
            # failure occurs through timeout
            await asyncio.wait_for(subprocess.wait(),
                                   timeout=ping_timeout)
            await self.set_info_and_report({'control_ping': 'on'})
            return
        except asyncio.TimeoutError:
            await self.set_info_and_report({'control_ping': 'off'})
            return

    async def probe_forever(self, cycle, ping_timeout, ssh_timeout):
        """
        runs forever, wait <cycle> seconds between 2 runs of probe()
        """
        while True:
            try:
                await self.probe(ping_timeout, ssh_timeout)
            except Exception:                           # pylint: disable=broad-except
                logger.exception("monitornodes oops 2")
            await asyncio.sleep(cycle)


class MonitorNodes:
    """
    monitor all nodes and reports their status to the sidecar service
    """

    def __init__(self, cmc_names, message_bus,
                 sidecar_url, cycle, verbose=False):
        self.cycle = cycle
        self.verbose = verbose

        # get miscell config
        self.ping_timeout = float(Config().value('networking', 'ping_timeout'))
        self.ssh_timeout = float(Config().value('networking', 'ssh_timeout'))
        self.log_period = float(Config().value('monitor', 'log_period'))

        # websockets
        self.reconnectable = \
            ReconnectableSidecar(sidecar_url, 'nodes')

        # the nodes part
        nodes = [Node(cmc_name, message_bus) for cmc_name in cmc_names]
        self.monitor_nodes = [
            MonitorNode(node=node, reconnectable=self.reconnectable,
                        verbose=verbose)
            for node in nodes]

    async def log(self):
        previous = 0
        while True:
            line = "".join([one_char_summary(mnode.info)
                            for mnode in self.monitor_nodes])
            current = self.reconnectable.counter
            delta = f"+ {current-previous}"
            line += f" {current} emits ({delta})"
            previous = current
            logger.warning(line)
            await asyncio.sleep(self.log_period)

    async def run_forever(self):
        logger.info(f"Starting nodes on {len(self.monitor_nodes)} nodes")
        return asyncio.gather(
            *[monitor_node.probe_forever(self.cycle,
                                         ping_timeout=self.ping_timeout,
                                         ssh_timeout=self.ssh_timeout)
              for monitor_node in self.monitor_nodes],
            self.reconnectable.keep_connected(),
            self.log(),
        )
